#
#   CritLoops by Paul Hsieh
#
#   Python test.
#
#   Copyright (c) 2012 Paul Hsieh
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#       Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#
#       Neither the name of CritLoops nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
#   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#   POSSIBILITY OF SUCH DAMAGE.
#

import os
import sys
import re
import string
import time
import random
from string import atoi

MSZDEFAULT = 2*1024*1024

class Bench:
  def __init__ (self, name):
    self.name = name;
    return

  def runtest (self, tc):
    startTime = 1000
    stopTime = 0
    count = 1
    DURATION = 3.0

    #  Calibrate the test so that it runs for a projected 
    #  time of DURATION seconds.  This calibration step can
    #  take between 0.4s and 0.8s to complete.

    while count > 0:
      startTime = time.clock()
      result = tc.subTest (count)
      stopTime = time.clock()
      if stopTime - startTime > 0.2: break
      count = count + count

    if 0 >= count:
      print tc.name() + ' is way too fast'
      return -1

    count = int ((DURATION * count) / (stopTime - startTime))
    if 0 >= count: count = 1

    #  Run the test 3 times.

    t0 = time.clock()
    result = tc.subTest (count)
    t1 = time.clock()
    result = tc.subTest (count)
    t2 = time.clock()
    result = tc.subTest (count)
    t3 = time.clock()

    t3 = t3 - t2
    t2 = t2 - t1
    t1 = t1 - t0

    if t3 > t2: t2,t3 = t3,t2
    if t1 >= t2: t3 = t2
    else: 
      if t1 >= t3: t3 = t1

    #  The official score of the test in units of iterations 
    #  per second.

    ips = count / t3

    # print '%s:\t%f iterations/second' % (tc.name(), ips)

    print "\tentry['%s']['sub']['%s'] = %f; // @CL@" % (self.name, tc.name(), ips)

    return ips

#  Basic string concatenation performance testing.  The amount
#  of concatenation is exaggerated versus real world applications
#  in order to simulate the most stressful corner cases.  Various
#  string implementation strategies, including length delimiting, 
#  copy on write, and fully referenced based collections can have 
#  a huge impact on the performance of string manipulation.

class stringConcatTest:
  def __init__ (self):
    self.TESTSTRING1 = 'CritLoop benchmark'
    return

  def subTest (self, count):
    a = self.TESTSTRING1
    c = 0

    for j in range (0, count):
      accum = ''
      for i in range (0, 250):
        accum = accum + a
        accum = accum + '!!'
      c = c + len (accum)
    return c

  def name (self):
    return 'String Concatenation'

#  Basic string searching performance testing.  String searching 
#  is commonly the counter balance to the most aggressive string
#  implementation strategies which emphasize concatenation speed.
#  It is difficult to be completely optimal on both.  The scenarios
#  run here also stress the algorithmic efficiency of string 
#  searching.

class stringSearchTest:
  def __init__ (self):
    self.TESTSTRING2 = 'Dot. date. datum. 123. Some more doubtful demonstration dummy data.'
    return

  def subTest (self, count):
    b = self.TESTSTRING2
    s = 'by'
    c = 0
    for j in range (0, count):
      for i in range (0, 250):
        c = c + b.find ('ummy')
        c = c + b.find (' data')
        for k in range (0, len (b)):
          if s.find (b[k]) > -1:
            c = c + k
            break
    return c

  def name (self):
    return 'String Scanning'

#  A standard 4x4 matrix multiplication.  Although matrix by vector
#  is more common, it can be considered a subset of this operation 
#  (the performance characteristics should be the same).  One matrix 
#  is assumed (runtime) constant and pre-transposed for better memory 
#  access.  The most inner loop is unrolled, as it is assumed that 
#  developers interested in this kernel would be concerned with
#  performance from the start.  The primary platform potential here 
#  should be to leverage the SIMD capabilities of modern processors.

class matrixMulGfx:

  def subTest (self, count):
    preTransposed = [ [ 0, 1, 0, 0 ],
                      [ 1, 0, 0, 0 ],
                      [ 0, 0, 1, 0 ],
                      [ 0, 0, 0, 1 ] ]

    target = [ [ 0, 0, 0, 0 ],
               [ 0, 0, 0, 0 ],
               [ 0, 0, 0, 0 ],
               [ 0, 0, 0, 0 ] ]

    source = [ [ 0, 0, 0, 0 ],
               [ 0, 0, 0, 0 ],
               [ 0, 0, 0, 0 ],
               [ 0, 0, 0, 0 ] ]

    for i in range (0, 4):
      for j in range (0, 4):
        source[i][j] = random.random()

    for k in range (0, count):
      for j in range (0, 1000):
        for i in range (0, 4):
          # Unrolled -- as opposed to FPUMark which was a loop
          target[i][0] = preTransposed[0][0] * source[i][0] + preTransposed[0][1] * source[i][1] + preTransposed[0][2] * source[i][2] + preTransposed[0][3] * source[i][3]
          target[i][1] = preTransposed[1][0] * source[i][0] + preTransposed[1][1] * source[i][1] + preTransposed[1][2] * source[i][2] + preTransposed[1][3] * source[i][3]
          target[i][2] = preTransposed[2][0] * source[i][0] + preTransposed[2][1] * source[i][1] + preTransposed[2][2] * source[i][2] + preTransposed[2][3] * source[i][3]
          target[i][3] = preTransposed[3][0] * source[i][0] + preTransposed[3][1] * source[i][1] + preTransposed[3][2] * source[i][2] + preTransposed[3][3] * source[i][3]
        for i in range (0, 4):
          source[i][0] = target[i][0]
          source[i][1] = target[i][1]
          source[i][2] = target[i][2]
          source[i][3] = target[i][3]
    return int (100.0 * target[3][3])

  def name (self):
    return 'Matrix Multiply GFX'

class mandelbrot:
  def __init__(self):
    self.MAX_ITERS = 2048

  def mandelPointTest (self, cr, ci):
    zr = cr
    zi = ci
    i = 0
    while i < self.MAX_ITERS and zr*zr + zi*zi < 4.0:
      tmp = zr * zr - zi * zi
      zi = 2 * zr * zi + ci
      zr = tmp + cr
      i = i + 1
    return i;

  def subTest (self, count):
    cr = -2.0
    ci = 2.0
    c = 0;
    for k in range (0, count):
      for i in range (0, 64):
        cr = (4.0 * (i & 63) / 64.0) - 2.0
        ci = 2.0 - (4.0 * (i & 63) / 64.0)
        c  = c + self.mandelPointTest (cr, ci)
    return c

  def name (self):
    return 'Mandelbrot'


#  These heap sorts use a not so common implementation of heap
#  sort that performs the major data swapping as a rotation 
#  operation.  This generally reduces data bandwidth by 50% over 
#  the common swap-based implementation.  It also peels off and 
#  simplifies the top-most case when walking down the heaps.  
#  Finally, it renders the inner-most comparison operation in 
#  such a way that it can be implemented as a conditional 
#  computation if the compiler and platform allow it.

class heapSortTest:
  def __init__ (self):
    self.len = 4099
    self.skip = 131

  def heapSortSiftDown (self, arr, top, last):
    maxIdx = top
    tmp = arr[top]
    maxIdx += maxIdx
    while last >= maxIdx:
      if maxIdx != last and arr[maxIdx] < arr[maxIdx + 1]:
        maxIdx = maxIdx + 1
      if tmp >= arr[maxIdx]: break
      arr[top] = arr[maxIdx]
      top = maxIdx
      maxIdx += maxIdx
    arr[top] = tmp

  def heapSortSiftDown0 (self, arr, last):
    if arr[1] > arr[0]:
      tmp = arr[1]
      arr[1] = arr[0]
      self.heapSortSiftDown (arr, 1, last)
    else:
      tmp = arr[0]
    arr[0] = arr[last]
    arr[last] = tmp

  def heapSort (self, arr):
    qty = len (arr)
    if qty > 1:
      i = int (qty / 2)
      qty = qty - 1
      while i > 0:
        self.heapSortSiftDown (arr, i, qty)
        i = i - 1
      i = qty
      while i > 0:
        self.heapSortSiftDown0 (arr, i)
        i = i - 1

  def subTest (self, count):
    arr = []
    j = 0
    for i in range (0, self.len):
      if j >= self.len: j = j - self.len
      arr = arr + [j]
      j = j + self.skip
    for k in range (0, count):
      self.heapSort (arr)
    return arr[self.len-1]

  def name (self):
    return 'Heapsort numbers'

#  Based on the "Collatz Sequence" open math problem.  The function 
#  computes how many steps are required before the terminal sequence 
#  4, 2, 1, 4 is encountered (the open conjecture is that this 
#  sequence is always reached from any starting condition.)

class collatzTest:
  def __init__ (self):
    self.COLLATZ_SMALL_TABLE_QTY = 16384;
    self.iter = [-1] * self.COLLATZ_SMALL_TABLE_QTY;

  def subTest (self, count):
    c = 0;
    for j in range (0, count):
      self.iter = [-1] * self.COLLATZ_SMALL_TABLE_QTY;
      mv = 0;
      mx = -1;
      for i in range (1, 113383):
        v = i
        curr = 0
        while v > 1:
          if v < self.COLLATZ_SMALL_TABLE_QTY and self.iter[v] > 0:
            curr += self.iter[v]
            break;
          if 0 != (v & 1):
            v = (3*v + 1) / 2
          else:
            v = v / 2
          curr += 1
        if curr > mv:
          mv = curr
          mx = i
        if i < self.COLLATZ_SMALL_TABLE_QTY:
          self.iter[i] = curr
      if mx != 106239 or mv != 223:
        print "Unexpected result |%d| -> %d" % (mx, mv)
      c += mv
    return c;

  def name (self):
    return 'Collatz Sequence';

#  These are the common "Stream" benchmarks for testing memory
#  bandwidth for floating point operations.  Basically it measures
#  the ability to sustain 1 or 2 streams read from memory and 1 
#  stream written while performing very light data dependent 
#  floating point operations.  Compilers have the opportunity to
#  detect that the arrays are non-aliasing and rerender the loops
#  into blocks for platforms which benefit from such 
#  transformations.  Alternatively, platforms/CPUs which can
#  handle multiple streams of simultaneous data access can remove
#  the need for better compilers.
# 
#  Note: these objects must be constructed with a parameter that 
#  guarantees that the arrays will not fit into any processor or
#  system cache.  The base size is 2097152.  The time results 
#  should be scalled accordingly.

class streamTest:
  def __init__ (self, sz):
    self.ARRAY_SIZE = sz
    self.a = [1.00000001] * sz
    self.b = [1.00000001] * sz
    self.c = [2.99999999] * sz

class streamCopy (streamTest):
  def subTest (self, count):
    for j in range (0, count):
      for i in range (0, self.ARRAY_SIZE): self.c[i] = self.a[i]
    return self.c[self.ARRAY_SIZE-1];

  def name (self):
    return 'Stream Copy'

class streamScale (streamTest):
  def subTest (self, count):
    for j in range (0, count):
      for i in range (0, self.ARRAY_SIZE): self.c[i] = 3.0 * self.a[i]
    return self.c[self.ARRAY_SIZE-1];

  def name (self):
    return 'Stream Scale'

class streamAdd (streamTest):
  def subTest (self, count):
    for j in range (0, count):
      for i in range (0, self.ARRAY_SIZE): self.c[i] = self.a[i] + self.b[i]
    return self.c[self.ARRAY_SIZE-1];

  def name (self):
    return 'Stream Add'

class streamTriad (streamTest):
  def subTest (self, count):
    for j in range (0, count):
      for i in range (0, self.ARRAY_SIZE): self.c[i] = self.a[i] + 3.0 * self.b[i]
    return self.c[self.ARRAY_SIZE-1];

  def name (self):
    return 'Stream Triad'

class latencyTest:
  def __init__ (self, sz):
    self.LLTEST_SIZE = sz
    self.d = [0] * sz
    for i in range (0, sz):
      f = i + 43711
      while f >= self.LLTEST_SIZE: f = f - self.LLTEST_SIZE
      self.d[i] = f

class latencyRAM (latencyTest):
  def subTest (self, count):
    k = 0
    for j in range (0, count):
      for i in range (0, 1024):
        k = self.d[k]

  def name (self):
    return 'Latency RAM'

class latencyL1Cache (latencyTest):
  def subTest (self, count):
    for i in range (0, 1024):
      self.d[i] = self.d[i] and 1023
    k = 0
    for j in range (0, count):
      for i in range (0, 1024):
        k = self.d[k]

  def name (self):
    return 'Latency L1 Cache'

if __name__=='__main__':

  try:
    import psyco
    psyco.full()
  except ImportError:
    pass

  print 'CritLoops for Python ('+sys.version+')'

  pname = 'Python ' + sys.version
  if len (sys.argv) > 1: pname = sys.argv[1]
  b = Bench (pname)

  print "\tentry['%s']['sub'] = new Array(); // @CL@" % (pname)

  fpuband  = b.runtest (matrixMulGfx ())
  fpulat   = b.runtest (mandelbrot ())
  strm     = b.runtest (stringConcatTest ())
  strs     = b.runtest (stringSearchTest ())
  flow     = b.runtest (heapSortTest ())
  flow2    = b.runtest (collatzTest ())
  msize = MSZDEFAULT
  memband1 = b.runtest (streamCopy (msize))
  memband2 = b.runtest (streamScale (msize))
  memband3 = b.runtest (streamAdd (msize))
  memband4 = b.runtest (streamTriad (msize))
  memlat1  = b.runtest (latencyRAM (msize))
  memlat2  = b.runtest (latencyL1Cache (msize))

  fpurate  = (fpuband * fpulat) ** 0.5
  strrate  = (strm * strs) ** 0.5
  flowrate = (flow * flow * flow2) ** (1/3.0)
  memrate  = ((memband1 * memband2 * memband3) ** (1/12.0)) * (memband4 * memlat1 * memlat2) ** (1/4.0)

  for x in ['std-rate','opt-rate']:
    print "\tentry['%s']['%s'] = new Array(); // @CL@" % (pname, x)
    print "\tentry['%s']['%s'][ 'Fpu'] = %.3f; // @CL@" % (pname, x, fpurate)
    print "\tentry['%s']['%s'][ 'Str'] = %.3f; // @CL@" % (pname, x, strrate)
    print "\tentry['%s']['%s']['Flow'] = %.3f; // @CL@" % (pname, x, flowrate)
    print "\tentry['%s']['%s'][ 'Mem'] = %.3f; // @CL@" % (pname, x, memrate)
    print "\tentry['%s']['%s'] = %.3f; // @CL@" % (pname, x[0:3], (fpurate * strrate * flowrate * memrate) ** (1/4.0))

  print ''

